//go:build !ignore_autogenerated
// +build !ignore_autogenerated

/*
Copyright The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

// Code generated by conversion-gen. DO NOT EDIT.

package v1beta1

import (
	unsafe "unsafe"

	storagemigrationv1beta1 "k8s.io/api/storagemigration/v1beta1"
	v1 "k8s.io/apimachinery/pkg/apis/meta/v1"
	conversion "k8s.io/apimachinery/pkg/conversion"
	runtime "k8s.io/apimachinery/pkg/runtime"
	storagemigration "k8s.io/kubernetes/pkg/apis/storagemigration"
)

func init() {
	localSchemeBuilder.Register(RegisterConversions)
}

// RegisterConversions adds conversion functions to the given scheme.
// Public to allow building arbitrary schemes.
func RegisterConversions(s *runtime.Scheme) error {
	if err := s.AddGeneratedConversionFunc((*storagemigrationv1beta1.StorageVersionMigration)(nil), (*storagemigration.StorageVersionMigration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_StorageVersionMigration_To_storagemigration_StorageVersionMigration(a.(*storagemigrationv1beta1.StorageVersionMigration), b.(*storagemigration.StorageVersionMigration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*storagemigration.StorageVersionMigration)(nil), (*storagemigrationv1beta1.StorageVersionMigration)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_storagemigration_StorageVersionMigration_To_v1beta1_StorageVersionMigration(a.(*storagemigration.StorageVersionMigration), b.(*storagemigrationv1beta1.StorageVersionMigration), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*storagemigrationv1beta1.StorageVersionMigrationList)(nil), (*storagemigration.StorageVersionMigrationList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_StorageVersionMigrationList_To_storagemigration_StorageVersionMigrationList(a.(*storagemigrationv1beta1.StorageVersionMigrationList), b.(*storagemigration.StorageVersionMigrationList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*storagemigration.StorageVersionMigrationList)(nil), (*storagemigrationv1beta1.StorageVersionMigrationList)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_storagemigration_StorageVersionMigrationList_To_v1beta1_StorageVersionMigrationList(a.(*storagemigration.StorageVersionMigrationList), b.(*storagemigrationv1beta1.StorageVersionMigrationList), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*storagemigrationv1beta1.StorageVersionMigrationSpec)(nil), (*storagemigration.StorageVersionMigrationSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_StorageVersionMigrationSpec_To_storagemigration_StorageVersionMigrationSpec(a.(*storagemigrationv1beta1.StorageVersionMigrationSpec), b.(*storagemigration.StorageVersionMigrationSpec), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*storagemigration.StorageVersionMigrationSpec)(nil), (*storagemigrationv1beta1.StorageVersionMigrationSpec)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_storagemigration_StorageVersionMigrationSpec_To_v1beta1_StorageVersionMigrationSpec(a.(*storagemigration.StorageVersionMigrationSpec), b.(*storagemigrationv1beta1.StorageVersionMigrationSpec), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*storagemigrationv1beta1.StorageVersionMigrationStatus)(nil), (*storagemigration.StorageVersionMigrationStatus)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_v1beta1_StorageVersionMigrationStatus_To_storagemigration_StorageVersionMigrationStatus(a.(*storagemigrationv1beta1.StorageVersionMigrationStatus), b.(*storagemigration.StorageVersionMigrationStatus), scope)
	}); err != nil {
		return err
	}
	if err := s.AddGeneratedConversionFunc((*storagemigration.StorageVersionMigrationStatus)(nil), (*storagemigrationv1beta1.StorageVersionMigrationStatus)(nil), func(a, b interface{}, scope conversion.Scope) error {
		return Convert_storagemigration_StorageVersionMigrationStatus_To_v1beta1_StorageVersionMigrationStatus(a.(*storagemigration.StorageVersionMigrationStatus), b.(*storagemigrationv1beta1.StorageVersionMigrationStatus), scope)
	}); err != nil {
		return err
	}
	return nil
}

func autoConvert_v1beta1_StorageVersionMigration_To_storagemigration_StorageVersionMigration(in *storagemigrationv1beta1.StorageVersionMigration, out *storagemigration.StorageVersionMigration, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_v1beta1_StorageVersionMigrationSpec_To_storagemigration_StorageVersionMigrationSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	if err := Convert_v1beta1_StorageVersionMigrationStatus_To_storagemigration_StorageVersionMigrationStatus(&in.Status, &out.Status, s); err != nil {
		return err
	}
	return nil
}

// Convert_v1beta1_StorageVersionMigration_To_storagemigration_StorageVersionMigration is an autogenerated conversion function.
func Convert_v1beta1_StorageVersionMigration_To_storagemigration_StorageVersionMigration(in *storagemigrationv1beta1.StorageVersionMigration, out *storagemigration.StorageVersionMigration, s conversion.Scope) error {
	return autoConvert_v1beta1_StorageVersionMigration_To_storagemigration_StorageVersionMigration(in, out, s)
}

func autoConvert_storagemigration_StorageVersionMigration_To_v1beta1_StorageVersionMigration(in *storagemigration.StorageVersionMigration, out *storagemigrationv1beta1.StorageVersionMigration, s conversion.Scope) error {
	out.ObjectMeta = in.ObjectMeta
	if err := Convert_storagemigration_StorageVersionMigrationSpec_To_v1beta1_StorageVersionMigrationSpec(&in.Spec, &out.Spec, s); err != nil {
		return err
	}
	if err := Convert_storagemigration_StorageVersionMigrationStatus_To_v1beta1_StorageVersionMigrationStatus(&in.Status, &out.Status, s); err != nil {
		return err
	}
	return nil
}

// Convert_storagemigration_StorageVersionMigration_To_v1beta1_StorageVersionMigration is an autogenerated conversion function.
func Convert_storagemigration_StorageVersionMigration_To_v1beta1_StorageVersionMigration(in *storagemigration.StorageVersionMigration, out *storagemigrationv1beta1.StorageVersionMigration, s conversion.Scope) error {
	return autoConvert_storagemigration_StorageVersionMigration_To_v1beta1_StorageVersionMigration(in, out, s)
}

func autoConvert_v1beta1_StorageVersionMigrationList_To_storagemigration_StorageVersionMigrationList(in *storagemigrationv1beta1.StorageVersionMigrationList, out *storagemigration.StorageVersionMigrationList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	out.Items = *(*[]storagemigration.StorageVersionMigration)(unsafe.Pointer(&in.Items))
	return nil
}

// Convert_v1beta1_StorageVersionMigrationList_To_storagemigration_StorageVersionMigrationList is an autogenerated conversion function.
func Convert_v1beta1_StorageVersionMigrationList_To_storagemigration_StorageVersionMigrationList(in *storagemigrationv1beta1.StorageVersionMigrationList, out *storagemigration.StorageVersionMigrationList, s conversion.Scope) error {
	return autoConvert_v1beta1_StorageVersionMigrationList_To_storagemigration_StorageVersionMigrationList(in, out, s)
}

func autoConvert_storagemigration_StorageVersionMigrationList_To_v1beta1_StorageVersionMigrationList(in *storagemigration.StorageVersionMigrationList, out *storagemigrationv1beta1.StorageVersionMigrationList, s conversion.Scope) error {
	out.ListMeta = in.ListMeta
	out.Items = *(*[]storagemigrationv1beta1.StorageVersionMigration)(unsafe.Pointer(&in.Items))
	return nil
}

// Convert_storagemigration_StorageVersionMigrationList_To_v1beta1_StorageVersionMigrationList is an autogenerated conversion function.
func Convert_storagemigration_StorageVersionMigrationList_To_v1beta1_StorageVersionMigrationList(in *storagemigration.StorageVersionMigrationList, out *storagemigrationv1beta1.StorageVersionMigrationList, s conversion.Scope) error {
	return autoConvert_storagemigration_StorageVersionMigrationList_To_v1beta1_StorageVersionMigrationList(in, out, s)
}

func autoConvert_v1beta1_StorageVersionMigrationSpec_To_storagemigration_StorageVersionMigrationSpec(in *storagemigrationv1beta1.StorageVersionMigrationSpec, out *storagemigration.StorageVersionMigrationSpec, s conversion.Scope) error {
	out.Resource = in.Resource
	return nil
}

// Convert_v1beta1_StorageVersionMigrationSpec_To_storagemigration_StorageVersionMigrationSpec is an autogenerated conversion function.
func Convert_v1beta1_StorageVersionMigrationSpec_To_storagemigration_StorageVersionMigrationSpec(in *storagemigrationv1beta1.StorageVersionMigrationSpec, out *storagemigration.StorageVersionMigrationSpec, s conversion.Scope) error {
	return autoConvert_v1beta1_StorageVersionMigrationSpec_To_storagemigration_StorageVersionMigrationSpec(in, out, s)
}

func autoConvert_storagemigration_StorageVersionMigrationSpec_To_v1beta1_StorageVersionMigrationSpec(in *storagemigration.StorageVersionMigrationSpec, out *storagemigrationv1beta1.StorageVersionMigrationSpec, s conversion.Scope) error {
	out.Resource = in.Resource
	return nil
}

// Convert_storagemigration_StorageVersionMigrationSpec_To_v1beta1_StorageVersionMigrationSpec is an autogenerated conversion function.
func Convert_storagemigration_StorageVersionMigrationSpec_To_v1beta1_StorageVersionMigrationSpec(in *storagemigration.StorageVersionMigrationSpec, out *storagemigrationv1beta1.StorageVersionMigrationSpec, s conversion.Scope) error {
	return autoConvert_storagemigration_StorageVersionMigrationSpec_To_v1beta1_StorageVersionMigrationSpec(in, out, s)
}

func autoConvert_v1beta1_StorageVersionMigrationStatus_To_storagemigration_StorageVersionMigrationStatus(in *storagemigrationv1beta1.StorageVersionMigrationStatus, out *storagemigration.StorageVersionMigrationStatus, s conversion.Scope) error {
	out.Conditions = *(*[]v1.Condition)(unsafe.Pointer(&in.Conditions))
	out.ResourceVersion = in.ResourceVersion
	return nil
}

// Convert_v1beta1_StorageVersionMigrationStatus_To_storagemigration_StorageVersionMigrationStatus is an autogenerated conversion function.
func Convert_v1beta1_StorageVersionMigrationStatus_To_storagemigration_StorageVersionMigrationStatus(in *storagemigrationv1beta1.StorageVersionMigrationStatus, out *storagemigration.StorageVersionMigrationStatus, s conversion.Scope) error {
	return autoConvert_v1beta1_StorageVersionMigrationStatus_To_storagemigration_StorageVersionMigrationStatus(in, out, s)
}

func autoConvert_storagemigration_StorageVersionMigrationStatus_To_v1beta1_StorageVersionMigrationStatus(in *storagemigration.StorageVersionMigrationStatus, out *storagemigrationv1beta1.StorageVersionMigrationStatus, s conversion.Scope) error {
	out.Conditions = *(*[]v1.Condition)(unsafe.Pointer(&in.Conditions))
	out.ResourceVersion = in.ResourceVersion
	return nil
}

// Convert_storagemigration_StorageVersionMigrationStatus_To_v1beta1_StorageVersionMigrationStatus is an autogenerated conversion function.
func Convert_storagemigration_StorageVersionMigrationStatus_To_v1beta1_StorageVersionMigrationStatus(in *storagemigration.StorageVersionMigrationStatus, out *storagemigrationv1beta1.StorageVersionMigrationStatus, s conversion.Scope) error {
	return autoConvert_storagemigration_StorageVersionMigrationStatus_To_v1beta1_StorageVersionMigrationStatus(in, out, s)
}
