/*
Copyright 2025 The Kubernetes Authors.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.
*/

package request

import "net/http"

const (
	MethodApply            string = "APPLY"
	MethodConnect          string = http.MethodConnect
	MethodCreate           string = "CREATE"
	MethodDelete           string = http.MethodDelete
	MethodDeleteCollection string = "DELETECOLLECTION"
	MethodGet              string = http.MethodGet
	MethodHead             string = http.MethodHead
	MethodList             string = "LIST"
	MethodPatch            string = http.MethodPatch
	MethodPost             string = http.MethodPost
	MethodProxy            string = "PROXY"
	MethodPut              string = http.MethodPut
	MethodUpdate           string = "UPDATE"
	MethodWatch            string = "WATCH"
	MethodWatchList        string = "WATCHLIST"
)
